#include <limits.h>
#include "bitfun.h"

#define BYTEBITS (sizeof(BYTE)*CHAR_BIT)

DWORD ReadBits(const BYTE *data, unsigned bitpos, unsigned bitcount)
{
    const unsigned bytepos = bitpos/BYTEBITS; // Ei nit tarvitse optimoida
    const unsigned offset = bitpos%BYTEBITS;  // kun kntj tekee sen itsekin
    const unsigned lastbytepos = (bitpos+bitcount-1)/BYTEBITS;
    const unsigned bitsinlastbyte = (bitpos+bitcount-1)%BYTEBITS + 1;

    DWORD result = data[bytepos] & ((BYTE)(~0)>>offset);

    if(lastbytepos == bytepos)
        return (result >> (BYTEBITS-bitsinlastbyte));
    
    {unsigned i;
    for(i = bytepos+1; i < lastbytepos; ++i)
    {
        result = (result<<BYTEBITS) | data[i];
    }}
    result =
        (result<<bitsinlastbyte) |
        (data[lastbytepos]>>(BYTEBITS-bitsinlastbyte));

    return result;
}

void WriteBits(BYTE *dest, unsigned bitpos, unsigned bitcount, DWORD data)
{
    const unsigned bytepos = bitpos/BYTEBITS;
    const unsigned offset = bitpos%BYTEBITS;
    const unsigned lastbytepos = (bitpos+bitcount-1)/BYTEBITS;
    const unsigned bitsinlastbyte = (bitpos+bitcount-1)%BYTEBITS + 1;

    if(bytepos == lastbytepos)
    {
        const BYTE mask =
            ((BYTE)(~0)<<(BYTEBITS-offset)) | ((BYTE)(~0)>>bitsinlastbyte);
        const unsigned databitsmask = (1<<bitcount)-1;
        dest[bytepos] =
            (dest[bytepos]&mask) |
            ((data&databitsmask)<<(BYTEBITS-bitsinlastbyte));
        return;
    }
    
    {const BYTE mask = (BYTE)(~0) >> bitsinlastbyte;
    dest[lastbytepos] =
        (dest[lastbytepos]&mask) | (BYTE)(data<<(BYTEBITS-bitsinlastbyte));
    data >>= bitsinlastbyte;
    }

    {unsigned i;
    for(i = lastbytepos-1; i > bytepos; --i)
    {
        dest[i] = (BYTE)(data);
        data >>= BYTEBITS;
    }}
    {const BYTE mask2 = (BYTE)(~0) << (BYTEBITS-offset);
    const unsigned databitsmask = (1<<(BYTEBITS-offset))-1;
    dest[bytepos] = (dest[bytepos]&mask2) | (data&databitsmask);
    }
}

void CopyBits(BYTE *dest, unsigned destbitpos, const BYTE *src, unsigned srcbitpos, unsigned count)
{
	while(count > 24)
	{
		WriteBits(dest, destbitpos, 24, ReadBits(src, srcbitpos, 24));
		
		destbitpos += 24;
		srcbitpos += 24;
		count -= 24;
	}
	if(count)
		WriteBits(dest, destbitpos, count, ReadBits(src, srcbitpos, count));
}

/* 0 = not equal, nonzero = equal */
int BitEqual(const BYTE *data, unsigned pos1, unsigned pos2, unsigned size)
{
	while(size > 24)
	{
		unsigned count = 24;
		if(ReadBits(data, pos1, count) != ReadBits(data, pos2, count))return 0;
		pos1 += count;
		pos2 += count;
		size -= count;
	}
	if(!size)return 1;
	return ReadBits(data, pos1, size) == ReadBits(data, pos2, size);
}
